#!/usr/bin/env python3

"""
__main__.py

Module for command-line execution of `panphylo`.
"""

# Import Python standard libraries
import argparse
import logging

# Import our library
import panphylo

# TODO: default output format to CSV, when stdout?
# TODO: add the ascertainment option

# TODO: remember to sync with README
def parse_args():
    """
    Parse command-line arguments and return them as a dictionary.
    """

    parser = argparse.ArgumentParser(description="Convert and manipulate phylodata.")
    parser.add_argument(
        "--input",
        type=str,
        default="-",
        help="Read input from *FILE*. If *FILE* is `-`, input will come from *stdin*.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="-",
        help="Write output to *FILE* instead of *stdout*. If *FILE* is `-`, output will go to *stdout* even if a non-textual format is specified.",
    )

    parser.add_argument(
        "-b",
        "--binarize",
        action="store_true",
        help="Binarizes the output. The specification on whether and how to add ascertainment correction is specified by the `--ascertainment` option.",
    )

    parser.add_argument(
        "-f",
        "--from",
        type=str,
        default="auto",
        choices=["auto", "tabular", "csv", "tsv", "nexus", "phylip"],
        help="Specify input format.",
    )
    parser.add_argument(
        "-e",
        "--encoding",
        type=str,
        default="auto",
        help="Character encoding for the input (use `auto` to detect).",
    )
    parser.add_argument(
        "-t",
        "--to",
        type=str,
        default="auto",
        choices=["auto", "csv", "tsv", "nexus", "phylip"],
        help="Specify output format.",
    )

    parser.add_argument(
        "--i-taxa",
        type=str,
        help="Input label, column, or name for taxa. Does not apply to all formats.",
    )
    parser.add_argument(
        "--i-char",
        type=str,
        help="Input label, column, or name for characters. Does not apply to all formats.",
    )
    parser.add_argument(
        "--i-state",
        type=str,
        help="Input label, column, or name for states. Does not apply to all formats.",
    )

    parser.add_argument(
        "--o-taxa",
        type=str,
        default="Taxon",
        help="Output label, column, or name for taxa. Does not apply to all formats.",
    )
    parser.add_argument(
        "--o-char",
        type=str,
        default="Character",
        help="Output label, column, or name for characters. Does not apply to all formats.",
    )
    parser.add_argument(
        "--o-state",
        type=str,
        default="State",
        help="Output label, column, or name for states. Does not apply to all formats.",
    )

    parser.add_argument(
        "--slug_taxa",
        type=str,
        default="simple",
        choices=["none", "simple", "full"],
        help="Level of slugging for taxa names.",
    )
    parser.add_argument(
        "--slug_chars",
        type=str,
        default="simple",
        choices=["none", "simple", "full"],
        help="Level of slugging for character names.",
    )

    parser.add_argument(
        "-v",
        "--verbosity",
        type=str,
        default="info",
        choices=["debug", "info", "warning", "error", "critical"],
        help="Set the logging level.",
    )

    args = parser.parse_args().__dict__

    # Decide on the right output function based on the output format or, if not
    # provided, on the extension
    if args["to"] == "auto":
        if not "." in args["to"]:
            raise ValueError(
                "Unable to detect output format; please specify it with `--to`."
            )

        extension = args["to"].split(".")[-1]
        if extension == "csv":
            args["to"] = "csv"
        elif extension == "tsv":
            args["to"] = "tsv"
        elif extension in ["nex", "nexus"]:
            args["to"] = "nexus"
        elif extension in ["phy", "phylip"]:
            args["to"] = "phylip"
        else:
            raise ValueError(
                "Unable to detect output format; please specify it with `--to`."
            )

    return args


def main():
    """
    Script entry point.
    """

    # Parse command line arguments and set the logger
    args = parse_args()

    level_map = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }
    logging.basicConfig(level=level_map[args["verbosity"]])

    # Read source data and detect the format if necessary
    source = panphylo.fetch_stream_data(args["input"], args["encoding"])
    if args["from"] == "auto":
        if source.strip().startswith("#NEXUS"):
            args["from"] = "nexus"
        elif re.search("$\s*\d+\s*\d+", source):
            args["from"] = "phylip"
        else:
            args["from"] = "tabular"

    # Convert to a string
    converted = panphylo.convert(source, args)

    # Write to the stream
    with panphylo.smart_open(args["output"], "w", encoding="utf-8") as handler:
        handler.write(converted.strip())
        handler.write("\n")


if __name__ == "__main__":
    main()



"""
Module with functions and methods for NEXUS files.

Parsing and writing of NEXUS files is currently done with a very simple,
string manipulation strategy.
"""

# TODO: allow to output taxa and character names between quotes, if necessary
# TODO: sort using assumptions (charset) if provided
# TODO: support comments (will need changes to the parser)

# Import Python libraries
import re
from enum import Enum, auto
from collections import defaultdict
from itertools import chain

# Import from local modules
from .internal import PhyloData
from .common import smart_open, indexes2ranges


def parse_nexus(source):
    """
    Parse the information in a NEXUS string.

    Parsing is currently done with a manually coded automaton that keeps track of
    state and buffers information until it can be used. It is not as advanced as
    some NEXUS parsing libraries, but this solution requires no additional package
    and can be gradually extended to our needs.
    """

    # Auxiliary state enumeration for the automaton
    class State(Enum):
        NONE = auto()
        OUT_OF_BLOCK = auto()
        IN_BLOCK = auto()

    # Data that will be filled during parsing; in the future, this could be expanded
    # to an actual class, with sanity checks etc.
    nexus_data = {
        "ntax": None,
        "nchar": None,
        "datatype": None,
        "missing": None,
        "gap": None,
        "symbols": None,
        "charstatelabels": {},
        "matrix": {},
        "charset": [],
    }

    buffer = ""
    block_name = ""
    state = State.NONE
    for idx, char in enumerate(source):
        # Extend buffer first and then process according to automaton state
        buffer += char

        if state == State.NONE:
            # Make sure we have a file that identifies as NEXUS
            if buffer.strip() == "#NEXUS":
                buffer = ""
                state = State.OUT_OF_BLOCK
        elif state == State.OUT_OF_BLOCK:
            # Make sure we have a block
            if char == ";":
                match = re.match(r"BEGIN\s+(.+)\s*;", buffer.upper().strip())
                if match:
                    block_name = match.group(1)
                    buffer = ""
                    state = State.IN_BLOCK
                else:
                    raise ValueError(f"Unable to parse NEXUS block at char {idx}.")
        elif state == State.IN_BLOCK:
            # Read all contents until we hit a semicolon, which will be processed individually
            if char == ";":
                # Check if we are at then of the block, otherwise process
                if re.sub(r"\s", "", buffer.upper().strip()) == "END;":
                    buffer = ""
                    state = State.OUT_OF_BLOCK
                else:
                    # Parse the command inside the block, which can be a single, simple
                    # line or a large subblock (like charstatelabels)
                    command = re.sub("\s+", " ", buffer.strip()).split()[0].upper()
                    if command == "DIMENSIONS":
                        ntax_match = re.search(r"NTAX\s*=\s*(\d+)", buffer.upper())
                        nchar_match = re.search(r"NCHAR\s*=\s*(\d+)", buffer.upper())
                        if ntax_match:
                            nexus_data["ntax"] = int(ntax_match.group(1))
                        if nchar_match:
                            nexus_data["nchar"] = int(nchar_match.group(1))
                    elif command == "FORMAT":
                        datatype_match = re.search(
                            r"DATATYPE\s*=\s*(\w+)", buffer.upper()
                        )
                        missing_match = re.search(r"MISSING\s*=\s*(.)", buffer.upper())
                        gap_match = re.search(r"GAP\s*=\s*(.)", buffer.upper())
                        symbols_match = re.search(
                            r"SYMBOLS\s*=\s*\"([^\"]+)\"", buffer.upper()
                        )
                        if datatype_match:
                            nexus_data["datatype"] = datatype_match.group(1)
                        if missing_match:
                            nexus_data["missing"] = missing_match.group(1)
                        if gap_match:
                            nexus_data["gap"] = gap_match.group(1)
                        if symbols_match:
                            # TODO: deal with space separated symbols
                            nexus_data["symbols"] = symbols_match.group(1)
                    elif command == "CHARSTATELABELS":
                        # Get each individual charstatelabel and parse it
                        # TODO: use a single parser for binary and multistate?
                        charstate_buffer = re.sub("\s+", " ", buffer.strip())
                        start_idx = charstate_buffer.find(
                            " ", charstate_buffer.find("CHARSTATELABELS")
                        )
                        for charstatelabel in charstate_buffer[start_idx:-1].split(","):
                            charstatelabel = re.sub("\s+", " ", charstatelabel.strip())
                            if "/" in charstatelabel:
                                # TODO: implement
                                raise ValueError("Not implemented")
                            else:
                                idx, charlabel = charstatelabel.split()
                                nexus_data["charstatelabels"][int(idx)] = charlabel
                    elif command == "MATRIX":
                        start_idx = buffer.find("MATRIX") + len("MATRIX")
                        for entry in buffer[start_idx + 1 : -1].strip().split("\n"):
                            entry = re.sub("\s+", " ", entry.strip())
                            taxon, vector = entry.split()
                            nexus_data["matrix"][taxon] = vector
                    elif command == "CHARSET":
                        # TODO: implement other syntaxes
                        # TODO: make sure it is performed only under the "assumption" block
                        match = re.search(
                            r"charset\s*(\w+)\s*=\s*(\d+)\s*-\s*(\d+)", buffer
                        )
                        if match:
                            charset_label, start, end = match.groups()
                            nexus_data["charset"].append(
                                {
                                    "charset": charset_label,
                                    "start": int(start),
                                    "end": int(end),
                                }
                            )

                    # Clean the buffer and continue
                    buffer = ""

    return nexus_data


def read_data_nexus(source, args):
    # Parse the NEXUS source
    nexus_data = parse_nexus(source)

    # Build internal representation
    # TODO: deal with multistate {}
    # TODO: transform all binary in multistate internal representation
    # TODO: this is currently handling only binary and needs charsets

    # Build inverse map from position in the alignment to the charset and
    # collect values
    alm2charset = {}
    for charset in nexus_data["charset"]:
        for idx in range(charset["start"], charset["end"] + 1):
            alm2charset[idx] = charset["charset"]

    values = defaultdict(set)
    for taxon, vector in nexus_data["matrix"].items():
        for idx, value in enumerate(vector):
            if value == nexus_data["missing"]:
                values[taxon, alm2charset[idx + 1]].add("?")
            elif value != "0":
                values[taxon, alm2charset[idx + 1]].add(
                    nexus_data["charstatelabels"][idx + 1]
                )

    phyd = PhyloData()
    for (taxon, character), value_set in values.items():
        for v in value_set:
            phyd.add_value(taxon, character, v)

    return phyd


def build_taxa_block(phyd):
    """
    Build a NEXUS TAXA block.
    """

    buffer = """
BEGIN TAXA;
    DIMENSIONS NTAX=%i;
    TAXLABELS
%s
    ;
END;""" % (
        len(phyd.taxa),
        "\n".join(["        %s" % taxon for taxon in sorted(phyd.taxa)]),
    )

    return buffer.strip()


# TODO: don't output charstatelabels values if all are binary
def build_character_block(phyd):
    # Express the actual labels only we have any state which is not a binary "0" or "1"
    # TODO: what about mixed data?
    states = sorted(set(chain.from_iterable(phyd.charvalues.values())))
    if tuple(states) == ("0", "1"):
        charstatelabels = [
            "        %i %s," % (charstate_idx + 1, character)
            for charstate_idx, (character, _) in enumerate(phyd.charvalues.items())
        ]
    else:
        charstatelabels = [
            "        %i %s /%s," % (charstate_idx + 1, character, " ".join(value_set))
            for charstate_idx, (character, value_set) in enumerate(
                phyd.charvalues.items()
            )
        ]

    # TODO: keeping the final comma in charstatelabels should be an option
    buffer = """
BEGIN CHARACTERS;
    DIMENSIONS NCHAR=%i;
    FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="%s";
    CHARSTATELABELS
%s
    ;

%s

END;""" % (
        len(phyd.charvalues),
        " ".join(phyd.symbols),
        "\n".join(charstatelabels),
        build_matrix_command(phyd),
    )

    return buffer.strip()


def build_matrix_command(phyd):
    """
    Build a NEXUS MATRIX command.
    """

    # Obtain the matrix and the maximum taxon length for formatting
    matrix = phyd.matrix
    taxon_length = max([len(entry["taxon"]) for entry in matrix])

    # Build buffer
    buffer = """
MATRIX
%s
;""" % (
        "\n".join(
            [
                "%s    %s" % (entry["taxon"].ljust(taxon_length), entry["vector"])
                for entry in matrix
            ]
        )
    )

    return buffer.strip()


def build_assumption_block(phyd):
    if not phyd.charset:
        return ""

    # Get the individual indexes first, and then build the string representation
    character_list = sorted(phyd.charvalues.keys())
    indexes = {
        charset: [character_list.index(char) + 1 for char in characters]
        for charset, characters in phyd.charset.items()
    }

    ##############
    # TODO; make sure it is sorted
    buffer = """
BEGIN ASSUMPTIONS;
%s
END;
    """ % (
        "\n".join(
            [
                "    CHARSET %s = %s;" % (charset, indexes2ranges(indexes[charset]))
                for charset in sorted(indexes)
            ]
        )
    )

    for charset, char_ranges in indexes.items():
        print(charset, char_ranges, indexes2ranges(char_ranges))

    return buffer


def build_nexus(phyd, args):
    # TODO: this only implements multistate
    # TODO: not rendering polymorphy

    components = [
        "#NEXUS",
        build_taxa_block(phyd),
        build_character_block(phyd),
        build_assumption_block(phyd),
    ]

    buffer = "\n\n".join([comp for comp in components if comp])

    return buffer



"""
Module with functions and methods for PHYLIP files.
"""

import re
from collections import defaultdict

# Import from local modules
from .common import smart_open
from .internal import PhyloData

# TODO: implement https://www.bioinformatics.org/sms/iupac.html
# TODO: currently only supporting non interleaved


def read_data_phylip(source, args):

    # Read raw data
    data = {}
    lines = source.strip().split("\n")

    match = re.search(r"(\d+) (\d+)", lines[0])
    ntax, nchar = int(match.group(1)), int(match.group(2))

    alms = [line.strip() for line in lines[1:] if line.strip()]
    for alm in alms:
        match = re.search(r"(\S+)\s+(.+)", alm)
        vector = match.group(2).replace(" ", "").upper()
        data[match.group(1)] = vector

    # Validate
    if len(data) != ntax:
        raise ValueError("Mismatch in number of taxa.")

    vector_lens = set([len(vector) for vector in data.values()])
    if len(vector_lens) != 1:
        raise ValueError("Mismatch in alignment lengths.")
    if list(vector_lens)[0] != nchar:
        raise ValueError("Alignment lengths differs from the one in the header.")

    # Add to PhyloData
    # TODO: add padding in the character names
    phyd = PhyloData()
    for taxon, vector in data.items():
        for char_idx, char_state in enumerate(vector):
            if char_state != "-":
                phyd.add_value(taxon, f"CHAR_{char_idx}", char_state)

    return phyd


# TODO: sharing matrix code in common with NEXUS, should move to PhyloData
def build_phylip(phyd, args):

    # Obtain the matrix and the maximum taxon length for formatting
    matrix = phyd.matrix
    taxon_length = max([len(entry["taxon"]) for entry in matrix])

    # Build buffer
    buffer = """
%i %i

%s
""" % (
        len(phyd.taxa),
        len(phyd.characters),
        "\n".join(
            [
                "%s    %s" % (entry["taxon"].ljust(taxon_length), entry["vector"])
                for entry in matrix
            ]
        ),
    )

    return buffer



"""
Module with functions and methods for tabular files.
"""

# Import Python libraries
import csv
import logging
from io import StringIO

# Import from local modules
from .common import smart_open, slug
from .internal import PhyloData


def detect_delimiter(source):
    """
    Detect the tabular dialect (e.g. CSV and TSV of a file).

    The detection is extremely simplified, based on frequency
    """

    lines = source.split("\n")
    commas = lines[0].count(",")
    tabs = lines[0].count("\t")
    logging.debug("Header has %i commas and %i tabs.", commas, tabs)
    if commas >= tabs:
        delimiter = ","
    else:
        delimiter = "\t"

    return delimiter


def _get_input_column_names(args, data):
    """
    Obtain column names, either provided or inferred.
    """

    # If the column names for taxa, characters, and values was not provided,
    # try to infer it; at the end, we make sure to check that they are all unique
    col_taxa = args.get("i-taxa", None)
    col_char = args.get("i-char", None)
    col_state = args.get("i-state", None)
    if not all([col_taxa, col_char, col_state]):
        logging.debug("Inferring column names.")

        # Get the keys we have and remove and column name already used
        columns = [
            col for col in data[0].keys() if col not in [col_taxa, col_char, col_state]
        ]

        # Obtain the taxa column among potential candidates, picking the first one
        for cand in [
            "taxon",
            "species",
            "language",
            "doculect",
            "manuscript",
            "witness",
        ]:
            for column in columns:
                if not col_taxa and cand in slug(column, "full"):
                    col_taxa = column

        # Obtain the char column among potential candidates, picking the first one
        for cand in ["character", "feature", "property", "position"]:
            for column in columns:
                if not col_char and cand in slug(column, "full"):
                    col_char = column

        # Obtain the value column among potential candidates, picking the first one
        for cand in ["state", "value", "observation", "cognate", "lesson", "reading"]:
            for column in columns:
                if not col_state and cand in slug(column, "full"):
                    col_state = column

    column_names = [col_taxa, col_char, col_state]
    if len(set(column_names)) < 3:
        raise AssertionError("Non-unique column names in %s", str(column_names))

    return column_names


# TODO: allow to prohibit column inference (should even be default?)
def read_data_tabular(source_str, delimiter, args):
    """
    Read data in tabular format.
    """

    # Read all data
    with StringIO(source_str) as handler:
        source = list(csv.DictReader(handler, delimiter=delimiter))
        logging.debug("Read %i entries from `%s`.", len(source), args["input"])

    # Infer column names
    col_taxa, col_char, col_vals = _get_input_column_names(args, source)

    # Build internal representation
    phyd = PhyloData()
    for entry in source:
        phyd.add_value(entry[col_taxa], entry[col_char], entry[col_vals])

    return phyd


def build_tabular(phyd, delimiter, args):
    # If the column names for taxa, characters, and states was not provided,
    # try to infer it; at the end, we make sure to check that they are all unique
    col_taxa = args.get("o-taxa", "Taxon")
    col_char = args.get("o-char", "Character")
    col_state = args.get("o-state", "State")

    # Build output data
    output = []
    for character in sorted(phyd.characters):
        for taxon in sorted(phyd.taxa):
            for value in sorted(phyd[taxon, character]):  # TODO: deal with missing
                output.append({col_taxa: taxon, col_char: character, col_state: value})

    # Write to an IO stream using the `csv` library, which
    # takes care of escapes etc.
    handler = StringIO()
    writer = csv.DictWriter(
        handler, delimiter=delimiter, fieldnames=[col_taxa, col_char, col_state]
    )
    writer.writeheader()
    writer.writerows(output)
    buffer = handler.getvalue()
    handler.close()

    # Fix issue with newlines as `\r\n`; as we were writing to a
    # StringIO(), this is the easiest way to do it
    buffer = buffer.replace("\r\n", "\n")

    return buffer



"""
Module with the class for the internal representation of data.
"""

# Import Python libraries
from collections import defaultdict, Counter
import enum
import string

# Import from local modules
from .common import unique_ids

# TODO; should have a general method for iterating over
#       characters in a sorter order
# TODO: have maximum taxon length here as well

# TODO: should have one for non-binary as well?
class BinaryObs(enum.Enum):
    """
    Auxiliary class for binary observations.

    The class allows to easily deal with missing data and gaps.
    """

    FALSE = enum.auto()
    TRUE = enum.auto()
    MISSING = enum.auto()
    GAP = enum.auto()


BINARY_OPS_MAP = {
    BinaryObs.FALSE: "0",
    BinaryObs.TRUE: "1",
    BinaryObs.MISSING: "?",
    BinaryObs.GAP: "-",
}


class PhyloData:
    """
    Class for the internal representation of phylogenetic data.
    """

    def __init__(self):
        """
        Initialize the class.
        """

        self.characters = set()
        self.charset = defaultdict(set)
        self.taxa = set()
        self.values = defaultdict(set)

    @property
    def charstates(self):
        """
        Return a sorted character states structure.

        Note that this method is not caching results; while it might involve
        more and unnecessary computations in some contexts, it does not affect
        our intended usage and makes the code easier to follow.

        :return: A dictionary with the character states.
        """

        # Collect with a counter first
        counter = defaultdict(Counter)
        for (_, character), states in self.values.items():
            counter[character].update({state for state in states if state != "?"})

        # Sort by frequency
        # TODO; add option to sort alphabetically
        # TODO: what if there are equal frequencies? Is it reproducible?
        charstates = {
            character: [state for state, _ in states.most_common()]
            for character, states in counter.items()
        }

        return charstates

    @property
    def char_cardinality(self) -> int:
        """
        Return the character cardinality.

        Character cardinality is defined as the number of states in the
        character(s) with the largest number of states.

        :return: An integer with the cardinality of the largest character.
        """

        return max([len(state_set) for state_set in self.charstates.values()])

    @property
    def matrix(self) -> list:
        """
        Build a sorted matrix from the internal representation.

        The matrix is returned as a list and is suitable for formats
        such as NEXUS and PHYLIP.

        :return: A sorted list with dictionaries carrying taxon and
            vector information.
        """

        # Build a sorted list with the matrix
        _matrix = defaultdict(str)
        symbols = self.symbols  # cache
        for character, state_set in self.charstates.items():
            for taxon in self.taxa:
                # TODO: assuming there is only one value per site!!! (value[0]), [None]
                state = self.values.get((taxon, character), [None])
                state = list(state)[0]
                if not state:
                    _matrix[taxon] += "-"
                elif state == "?":
                    _matrix[taxon] += "?"
                else:
                    _matrix[taxon] += symbols[state_set.index(state)]

        ret = [{"taxon": taxon, "vector": vector} for taxon, vector in _matrix.items()]
        ret = sorted(ret, key=lambda e: e["taxon"])

        return ret

    @property
    def symbols(self):
        """
        Return a colloction of unique symbols for representing states.

        The length of the vector will follow the cardinality of the data
        (i.e., `self.char_cardinality`).

        :return: A list with the characters for representing the states.
        """

        return [ch for ch in string.digits + string.ascii_uppercase][
            : self.char_cardinality
        ]

    def slug_taxa(self, level="simple"):
        """
        Slug taxa labels, making sure uniqueness of IDs is preserved.
        """

        # If level is "none", we only make sure the IDs are now unique

        # Build map with unique ids, update self.taxa, and update self.values
        slug_map = {
            source: target
            for source, target in zip(self.taxa, unique_ids(self.taxa, level))
        }

        self.taxa = set(slug_map.values())

        self.values = {
            (slug_map[taxon], character): values
            for (taxon, character), values in self.values.items()
        }

    def slug_characters(self, level="simple"):
        """
        Slug character labels, making sure uniqueness of IDs is preserved.
        """

        # Build map with unique ids, update self.characters, and update self.values
        slug_map = {
            source: target
            for source, target in zip(
                self.characters, unique_ids(self.characters, level)
            )
        }

        self.characters = set(slug_map.values())

        self.values = {
            (taxon, slug_map[character]): values
            for (taxon, character), values in self.values.items()
        }

    def binarize(self):
        # TODO: collect assumptions

        # For each taxon, collect a map of which charvalues are observed; this does not modify the
        # internal properties
        charvalues = self.charstates  # cache
        binary_values = {}
        for taxon in self.taxa:
            for character, values in self.charstates.items():
                obs = self.values.get((taxon, character), None)

                if not obs:
                    binary_values[taxon, character] = [
                        BinaryObs.GAP for val in charvalues[character]
                    ]
                else:
                    if tuple(obs) == "?":
                        binary_values[taxon, character] = [
                            BinaryObs.MISSING for val in charvalues[character]
                        ]
                    else:
                        binary_values[taxon, character] = [
                            BinaryObs.TRUE if val in obs else BinaryObs.FALSE
                            for val in charvalues[character]
                        ]

        # Build a new phylogenetic data structure, adding the new characters one by one
        bin_phyd = PhyloData()
        for (taxon, character), value in binary_values.items():
            for obs, value_name in zip(value, charvalues[character]):
                # Add ascertainment
                # TODO: allow something different from "0"?
                # TODO: should check for the name of the character?
                bin_phyd.add_value(
                    taxon, f"{character}_ASCERTAINMENT", "0", charset=character
                )

                # TODO: confirm if it is right to skip over gaps and all
                if obs != BinaryObs.GAP:
                    bin_phyd.add_value(
                        taxon,
                        f"{character}_{value_name}",
                        BINARY_OPS_MAP[obs],
                        charset=character,
                    )

        return bin_phyd

    # TODO: move to __setitem__? it is actually an "add"
    def add_value(self, taxon, character, value, charset=None):
        """
        Add a value to a taxon, character pair.

        Nota the (taxon, character) pairs can carry more than
        one value, so that values are actually stored in sets
        """

        self.values[taxon, character].add(value)
        self.taxa.add(taxon)
        self.characters.add(character)

        if charset:
            self.charset[charset].add(character)

    def __getitem__(self, key):
        return self.values[key]

    # TODO: add total number of values
    def __repr__(self):
        return f"PhyloData with {len(self.taxa)} taxa and {len(self.characters)} characters."



"""
Module with common and reusable functions.
"""

# Import Python libraries
import re
import string
import unidecode
import itertools
import logging
import chardet

import sys
import contextlib

# TODO: for slug, consider that The following symbols/caracters are not allowed in taxa names to ensure Newick
#       compatibility: (space), (semicolon), (colon), (comma), (parentheses), (single quote)


@contextlib.contextmanager
def smart_open(filename: str, mode: str = "r", *args, **kwargs):
    """
    Open files and i/o streams transparently.

    Code originally from https://stackoverflow.com/a/45735618.
    """

    if filename == "-":
        if "r" in mode:
            stream = sys.stdin
        else:
            stream = sys.stdout
        if "b" in mode:
            fh = stream.buffer
        else:
            fh = stream
        close = False
    else:
        fh = open(filename, mode, *args, **kwargs)
        close = True

    try:
        yield fh
    finally:
        if close:
            try:
                fh.close()
            except AttributeError:
                pass


def fetch_stream_data(input, encoding="auto") -> str:
    """
    Read the input data as a string.

    The function takes care of handling input from both stdin and
    files, decoding the stream of bytes according to the user-specified
    character encoding (including automatic detection if necessary).

    :param args: The arguments carrying all options, either converted
        from command-line arguments, built from an interface, or
        as a dictionary.
    :return: A string with the full source for the data, encoded
        according to the specified charset encoding.
    """

    # Fetch all input as a sequence of bytes, so that we don't consume stdout
    # and can still run autodetection on format and encoding
    with smart_open(input, "rb") as handler:
        logging.debug("Reading contents from `%s`.", input)
        raw_source = handler.read()

        # Detect encoding if necessary, building a string
        if encoding != "auto":
            logging.debug("Using `%s` character encoding.", encoding)
        else:
            detect = chardet.detect(raw_source)
            encoding = detect["encoding"]
            logging.debug(
                "Encoding detected as `%s` (confidence: %.2f)",
                detect["encoding"],
                detect["confidence"],
            )

        source = raw_source.decode(encoding)

    return source


# TODO: allow more configurations
# TODO: document levels
def slug(label, level):
    """
    Return a slugged version of a label.
    """

    if level == "none":
        pass
    elif level == "simple":
        label = unidecode.unidecode(label)
        label = re.sub("\s+", "_", label.strip())
        label = "".join(
            [
                char
                for char in label
                if char in string.ascii_letters + string.digits + "-_"
            ]
        )
    elif level == "full":
        label = unidecode.unidecode(label)
        label = re.sub("\s+", " ", label.strip())
        label = label.lower()
        label = "".join([char for char in label if char in string.ascii_letters])

    return label


# TODO: add different methods of slug, mapping to slug()
def unique_ids(labels, level):
    def _label_iter():
        """
        Custom internal label iterator.

        -> "a", "b", ..., "aa", "ab", ..., "zz", "aaa", "aab", ...
        """
        for length in itertools.count(1):
            for chars in itertools.product(string.ascii_lowercase, repeat=length):
                yield "-" + "".join(chars)

    # Slugify all labels
    slugged = [slug(label, level) for label in labels]

    # Build a corresponding list with the count of previous occurrences
    # of the same value
    loc_counts = [slugged[:idx].count(value) for idx, value in enumerate(slugged)]

    # Build a dictionary of suffixes using the maximum count
    # NOTE: This might look as an overkill, but will allow to easily
    #       adapt other models in the future
    label_iter = _label_iter()
    suffix = {count: next(label_iter) for count in range(max(loc_counts) + 1)}

    # Build the new list, adding the suffix if there is more than one
    # occurrence overall, and return
    # TODO: the overall count could be cached with a collections.Counter
    unique_slug_labels = [
        f"{label}{suffix[loc_count]}" if slugged.count(label) > 1 else label
        for label, loc_count in zip(slugged, loc_counts)
    ]

    return unique_slug_labels


def indexes2ranges(indexes, string=True):
    """
    Transforms a list of indexes into a range representation.

    This function is used for building NEXUS-like assumption blocks,
    especially for binarized data. Given a list such as `[1, 2, 3, 5, 8, 9]`
    it will return either a structural or a string representation of
    the ranges involved, such as `"1-3, 5, 8-9"`.
    """

    indexes = sorted(indexes)

    # Collect ranges as a list of tuples
    ranges = []
    start, end = None, None
    for idx in indexes:
        if not start:
            start = idx
        elif not end:
            if idx > start + 1:
                ranges.append((start, start))
                start = idx
            else:
                end = idx
        else:
            if idx == end + 1:
                end = idx
            else:
                ranges.append((start, end))
                start, end = idx, None

    # We finish with whatever is in the pile; note that we add `indexes[-1]`
    # and not `end`, dealing with single sites
    ranges.append((start, indexes[-1]))

    # Build output
    if not string:
        ret = tuple(ranges)
    else:
        ret = ", ".join(
            [
                "%i-%i" % (start, end) if start != end else str(start)
                for (start, end) in ranges
            ]
        )

    return ret



"""
setup.py for the `panphylo` library and tool

Requirements are listed in `requirements.txt`.
"""


# Import Python standard libraries
from setuptools import setup, find_packages
import pathlib

# The directory containing this file
LOCAL_PATH = pathlib.Path(__file__).parent

# The text of the README file
README_FILE = (LOCAL_PATH / "README.md").read_text(encoding="utf-8")

# Load requirements, so they are listed in a single place
with open("requirements.txt", encoding="utf-8") as fp:
    install_requires = [dep.strip() for dep in fp.readlines()]

# This call to setup() does all the work
# TODO: expand classifiers and keywords
setup(
    author="Tiago Tresoldi",
    author_email="tiago.tresoldi@lingfil.uu.se",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
    ],
    description="A tool for converting and manipulating phylogenetic data",
    entry_points={"console_scripts": ["panphylo=panphylo.__main__:main"]},
    extras_require={
        "dev": ["black", "flake8", "twine", "wheel"],
        "test": ["pytest"],
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords=["phylogenetics", "nexus", "phylip"],
    license="MIT",
    long_description=README_FILE,
    long_description_content_type="text/markdown",
    name="panphylo",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.6",
    test_suite="tests",
    tests_require=[],
    url="https://github.com/tresoldi/panphylo",
    version="0.1",  # remember to sync with __init__.py
    zip_safe=False,
)



"""
panphylo __init__.py
"""

# Version of the panphylo package
__version__ = "0.1"  # remember to sync in setup.py
__author__ = "Tiago Tresoldi"
__email__ = "tiago.tresoldi@lingfil.uu.se"

# Import from local modules
from .common import indexes2ranges, smart_open, fetch_stream_data
from .nexus import read_data_nexus, build_nexus
from .phylip import read_data_phylip, build_phylip
from .tabular import detect_delimiter, read_data_tabular, build_tabular


# Dispatch the different reading methods
def convert(source, args):

    # Dispatch to the right reading method
    if args["from"] == "tabular":
        phyd = read_data_tabular(source, detect_delimiter(source), args)
    elif args["from"] == "csv":
        phyd = read_data_tabular(source, ",", args)
    elif args["from"] == "tsv":
        phyd = read_data_tabular(source, "\t", args)
    elif args["from"] == "nexus":
        phyd = read_data_nexus(source, args)
    elif args["from"] == "phylip":
        phyd = read_data_phylip(source, args)

    # Perform all requested data manipulations
    phyd.slug_taxa(args.get("slug_taxa", "none"))
    phyd.slug_characters(args.get("slug_chars", "none"))
    if args.get("binarize", False):
        phyd = phyd.binarize()

    # Write converted data in the requested format; note that the command-line
    # handling should have taken care of replacing the "auto" value for
    # autodetecting the output
    # TODO: adapt to have them returning strings, and writing with a separate function
    if args["to"] == "csv":
        converted = build_tabular(phyd, ",", args)
    elif args["to"] == "tsv":
        converted = build_tabular(phyd, "\t", args)
    elif args["to"] == "nexus":
        converted = build_nexus(phyd, args)
    elif args["to"] == "phylip":
        converted = build_phylip(phyd, args)

    return converted


# Build namespace
__all__ = ["convert", "indexes2ranges", "smart_open", "fetch_stream_data"]
